#include <stdio.h>

#include "strat.hpp"

#define max(a,b) ((a) > (b) ? (a) : (b))


strat::strat(int numdecks, int r, int m) {

  shoe_reset(numdecks);
  rules = r;
  maxsplits = m;

}


void strat::shoe_reset(int numdecks) {

  int i;

  for (i=1; i<10; i++) cards[i] = numdecks*4;

  cards[10] = numdecks*4*4;

  numcards = 52*numdecks;

}


inline void strat::shoe_remove(int value) {

  cards[value]--;
  numcards--;

}


inline void strat::shoe_add(int value) {

  cards[value]++;
  numcards++;

}


inline float strat::prob(int value) {

  return ((float)cards[value]) / numcards;

}



int strat::insurance_decide() {

  if (prob(10) > 1.0/3.0) return 1;
  else return 0;

}



/* resprobs[0] : WIN   resprobs[1] : LOSE   resprobs[2] : PUSH */

void strat::split_probs(float probmult, float *probs, float *resprobs, hand *th, hand *dh) {

  int a,i;
  float tresprobs[4];
  float expect=0;

  a = th->single(0);

  for (i=1; i<11; i++) {
    th->changelast(i);

    decide_aux(prob(i)*probmult, ACTION_DOUBLE | (i == a ? ACTION_SPLIT : 0), &tresprobs[0], th, dh);

    expect += max(tresprobs[3], max(tresprobs[2], max(tresprobs[1], tresprobs[0])));
  }

  resprobs[0] += 2 * expect;

  th->changelast(a);

}





/* resprobs[0] : WIN   resprobs[1] : LOSE   resprobs[2] : PUSH */

void strat::double_probs(float probmult, float *probs, float *resprobs, hand *th, hand *dh) {

  float tprob;
  int i,j,hv;

  th->addcard(1);

  for (i=1; i<11; i++) {
    tprob = prob(i)*probmult;

    th->changelast(i);

    hv = th->handval();

    if (hv > 21) {
      resprobs[1] += tprob * 2;
      continue;
    }

    resprobs[0] += probs[6] * tprob * 2;

    // Since you get to keep your double-down on dealer blackjack:

    resprobs[1] += probs[5] * tprob;

    for (j=17; j<22; j++) {

      if (j < hv) resprobs[0] += probs[j-17] * tprob * 2;
      else if (j > hv) resprobs[1] += probs[j-17] * tprob * 2;
      else if (j == hv) resprobs[2] += probs[j-17] * tprob * 2;

    }

  }

  th->remlast();

}





/* resprobs[0] : WIN   resprobs[1] : LOSE   resprobs[2] : PUSH */

void strat::hit_probs(float probmult, float *probs, float *resprobs, hand *th, hand *dh) {

  float st_probs[3], ht_probs[3];
  float tprobmult;
  int i,j;

  if (th->handval() >= 21) {
    resprobs[1] += probmult;
    return;
  }

  if (probmult < .000001) return;

  for (i=1; i<11; i++) {

    tprobmult = probmult * prob(i);

    th->addcard(i);

    if (th->handval() > 21) {
      resprobs[1] += tprobmult;
      th->remlast();
      continue;
    }

    for (j=0; j<3; j++) {
      st_probs[j] = 0.0;
      ht_probs[j] = 0.0;
    }

    stand_probs(tprobmult, probs, st_probs, th, dh);

    shoe_remove(i);
    hit_probs(tprobmult, probs, ht_probs, th, dh);
    shoe_add(i);

    th->remlast();

    if (st_probs[0] - st_probs[1] > ht_probs[0] - ht_probs[1]) {
      resprobs[0] += st_probs[0];
      resprobs[1] += st_probs[1];
      resprobs[2] += st_probs[2];
    } else {
      resprobs[0] += ht_probs[0];
      resprobs[1] += ht_probs[1];
      resprobs[2] += ht_probs[2];
    }

  }

}




/* resprobs[0] : WIN   resprobs[1] : LOSE   resprobs[2] : PUSH */

void strat::stand_probs(float probmult, float *probs, float *resprobs, hand *th, hand *dh) {

  int hv, i;

  hv = th->handval();

  if (hv < 17) {
    resprobs[0] += probmult * probs[6];
    resprobs[1] += probmult - (probmult * probs[6]);
  } else if (hv > 21) {
    resprobs[1] += probmult;
  } else if (hv == 21) {
    if (th->num() == 2) {
      resprobs[2] += probmult * probs[5];
      // We don't pay 3:2 if this natural came from a split
      if (currsplit) {
        resprobs[0] += (probmult - (probmult * probs[5]));
      } else {
        resprobs[0] += (probmult - (probmult * probs[5])) * 1.5;
      }
    } else {
      resprobs[1] += probmult * probs[5];
      resprobs[2] += probmult * probs[4];
      resprobs[0] += probmult - (probmult * probs[5]) - (probmult * probs[4]);
    }
  } else {
    resprobs[0] += probmult * probs[6];
    resprobs[1] += (probmult * probs[5]) + (probmult * probs[4]);

    for (i=17; i<21; i++) {
      if (i == hv) resprobs[2] += probmult * probs[i-17];
      else if (i < hv) resprobs[0] += probmult * probs[i-17];
      else if (i > hv) resprobs[1] += probmult * probs[i-17];
    }
  }

}


/* probs[0] : 17   probs[1] : 18   probs[2] : 19
   probs[3] : 20   probs[4] : 21
   probs[5] : BJ   probs[6] : BUST
 */

void strat::dealer_probs(float *probs, hand *dh) {

  int i;

  for (i=0; i<7; i++) probs[i] = 0.0;

  dealer_probs_aux(1.0, probs, dh);

}


void strat::dealer_probs_aux(float probmult, float *probs, hand *dh) {

  int i, hv, sft;
  float tprobmult;

  if (probmult < .00001) return;

  hv = dh->handval();
  sft = dh->issoft();

  if (hv >= 22) {
    probs[6] += probmult;
    return;
  } else if (hv == 21 && dh->num() == 2) {
    probs[5] += probmult;
    return;
  } else if (hv >= 17) {
    if (hv == 17 && sft && (rules & RULE_H17)) {
      // do nothing - we hit soft 17
    } else {
      probs[hv-17] += probmult;
      return;
    }
  }

  // FIXME: Possible algorithmic improvement:
  // If, for instance, we bust on a 6 we know that
  // everything higher will also bust. We could simply add the
  // rest to probs[6] (BUST) after the first bust occurs.

  // Just need to add anything at all
  dh->addcard(1);

  for (i=1; i<11; i++) {

    tprobmult = probmult * prob(i);

    if (tprobmult == 0.0) continue;

    dh->changelast(i);

    shoe_remove(i);

    dealer_probs_aux(tprobmult, probs, dh);

    shoe_add(i);

  }

  dh->remlast();

}




void strat::decide(float *expect, int a, int b, int up) {

  int allow;
  hand th(a, b);
  hand dh(up);

  expect[0] = expect[1] = expect[2] = expect[3] = 0.0;

  allow = ACTION_DOUBLE;
  if (a == b) allow |= ACTION_SPLIT;

  if (prob(a) == 0.0 || prob(b) == 0.0 || prob(up) == 0.0) {
    // FIXME: Better error handling
    fprintf(stderr, "BAD\n");
    exit(0);
  }

  shoe_remove(a);
  shoe_remove(b);
  shoe_remove(up);

  decide_aux(1.0, allow, expect, &th, &dh);

  currsplit = 0;

  shoe_add(up);
  shoe_add(b);
  shoe_add(a);
  
}



// ACTION_HIT and ACTION_STAND are implied

// Notice at this point that the cards in th and dh have
// been removed from the shoe so we don't need to worry about them

// The splitting implementation needs to consider these notes:

// RULE_DAS/RULE_RSA
// Be sure not to pay out 3:2 after splitting aces
// Only get 1 card after splitting aces
// Maxsplits gets enforced

// expect[0] : stand    expect[1] : hit
// expect[2] : double   expect[3] : split

void strat::decide_aux(float probmult, int allow, float *expect, hand *th, hand *dh) {

  float dprobs[7];
  float resprobs[3];

  if (probmult < .000001) return;

  expect[0] = expect[1] = expect[2] = expect[3] = -1000.0;

  dealer_probs(&dprobs[0], dh);

  resprobs[0] = resprobs[1] = resprobs[2] = 0.0;
  stand_probs(probmult, &dprobs[0], &resprobs[0], th, dh);
  expect[0] = resprobs[0] - resprobs[1];
  //printf("S PROBS: %g %g %g\n", resprobs[0], resprobs[1], resprobs[2]);

  if (currsplit == 0 || th->single(0) != 1) {
    resprobs[0] = resprobs[1] = resprobs[2] = 0.0;
    hit_probs(probmult, &dprobs[0], &resprobs[0], th, dh);
    expect[1] = resprobs[0] - resprobs[1];
    //printf("H PROBS: %g %g %g\n", resprobs[0], resprobs[1], resprobs[2]);

    if (allow & ACTION_DOUBLE) {
      resprobs[0] = resprobs[1] = resprobs[2] = 0.0;
      double_probs(probmult, &dprobs[0], &resprobs[0], th, dh);
      expect[2] = resprobs[0] - resprobs[1];
      //printf("D PROBS: %g %g %g\n", resprobs[0], resprobs[1], resprobs[2]);
    }
  }

  if (allow & ACTION_SPLIT && currsplit < maxsplits) {
    currsplit++;
    resprobs[0] = resprobs[1] = resprobs[2] = 0.0;
    split_probs(probmult, &dprobs[0], &resprobs[0], th, dh);
    expect[3] = resprobs[0] - resprobs[1];
    //printf("SP PROBS: %g %g %g\n", resprobs[0], resprobs[1], resprobs[2]);
  }

}
